Package: flowSpecs
Version: 1.24.0
Date: 2023-04-05
Type: Package
Title: Tools for processing of high-dimensional cytometry data
biocViews: Software,CellBasedAssays,DataRepresentation,ImmunoOncology,
        FlowCytometry,SingleCell,Visualization,Normalization,DataImport
Authors@R: person('Jakob', 'Theorell', email = "jakob.theorell@ki.se",
        role = c("aut", "cre"))
Description: This package is intended to fill the role of conventional cytometry
        pre-processing software, for spectral decomposition, transformation, 
        visualization and cleanup, and to aid further downstream analyses, such
        as with DepecheR, by enabling transformation of flowFrames and flowSets
        to dataframes. Functions for flowCore-compliant automatic 
        1D-gating/filtering are in the pipe line. 
        The package name has been chosen both as it will deal with spectral
        cytometry and as it will hopefully give the user a nice pair of
        spectacles through which to view their data.
BugReports: https://github.com/jtheorell/flowSpecs/issues
License: MIT + file LICENSE
Encoding: UTF-8
LazyData: false
RoxygenNote: 7.2.3
Depends: R (>= 4.0)
Imports: ggplot2 (>= 3.1.0), BiocGenerics (>= 0.30.0), BiocParallel (>=
        1.18.1), Biobase (>= 2.48.0), reshape2 (>= 1.4.3), flowCore (>=
        1.50.0), zoo (>= 1.8.6), stats (>= 3.6.0), methods (>= 3.6.0)
Suggests: testthat, knitr, rmarkdown, BiocStyle, DepecheR
VignetteBuilder: knitr
git_url: https://git.bioconductor.org/packages/flowSpecs
git_branch: RELEASE_3_22
git_last_commit: 2da6fa3
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.22
Date/Publication: 2025-10-29
NeedsCompilation: no
Packaged: 2025-10-30 04:00:17 UTC; biocbuild
Author: Jakob Theorell [aut, cre]
Maintainer: Jakob Theorell <jakob.theorell@ki.se>

# Generated by roxygen2: do not edit by hand

export(arcTrans)
export(corrMatCreate)
export(correctUnmix)
export(flowSet2LongDf)
export(oneVsAllPlot)
export(peakNorm)
export(specMatCalc)
export(specUnmix)
importFrom(Biobase,"pData<-")
importFrom(Biobase,pData)
importFrom(BiocGenerics,colnames)
importFrom(BiocGenerics,ncol)
importFrom(BiocGenerics,nrow)
importFrom(BiocParallel,bplapply)
importFrom(flowCore,"exprs<-")
importFrom(flowCore,"parameters<-")
importFrom(flowCore,"sampleNames<-")
importFrom(flowCore,exprs)
importFrom(flowCore,flowSet)
importFrom(flowCore,fsApply)
importFrom(flowCore,keyword)
importFrom(flowCore,pData)
importFrom(flowCore,parameters)
importFrom(flowCore,sampleNames)
importFrom(ggplot2,aes)
importFrom(ggplot2,element_blank)
importFrom(ggplot2,element_line)
importFrom(ggplot2,facet_wrap)
importFrom(ggplot2,geom_hex)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,ggsave)
importFrom(ggplot2,scale_fill_gradientn)
importFrom(ggplot2,theme)
importFrom(ggplot2,xlab)
importFrom(ggplot2,ylab)
importFrom(methods,setGeneric)
importFrom(methods,setMethod)
importFrom(reshape2,melt)
importFrom(stats,density)
importFrom(stats,lsfit)
importFrom(stats,mad)
importFrom(stats,median)
importFrom(stats,quantile)
importFrom(stats,var)
importFrom(zoo,rollmean)

#' Efficient inverse hyperbolic cosine transformation
#'
#' This is a simple wrapper function for the base asinh function, that is useful
#' for flowFrames and flowSets. It also allows for reversing the transformation
#' with the argument "unTrans".
#' @importFrom BiocGenerics colnames
#' @importFrom flowCore exprs fsApply
#' @param flowObj The fcs object to be transformed. Both flowFrames and flowSets
#' are accepted.
#' @param transNames The variables that should be normalized.
#' @param transCoFacs This value or vector of values define the values for the
#' transformation during the normalization. In the "default" case, the function
#' defines the object as a CyTOF object if >5 percent of the values are 0, and
#' applies the transformation value 5. Otherwise, the value 400 is applied.
#' @param unTrans If the reverse action should be taken, i.e. if an already
#' transformed dataset should be un-transformed. NB! It is of great importance
#' that the same transformation factors are used!
#' @return A flow object containing the transformed data, and with all metadata
#' left untouched.
#' @examples
#' # Import some data and the spectral matrix. The latter can be generated using
#' # specMatCalc
#' data(fullPanel)
#' data(specMat)
#'
#' fullPanelUnmixed <- specUnmix(fullPanel, specMat)
#'
#' # Identify the columns that should be transformed
#' colnames(fullPanelUnmixed)
#' # The time and scatter parameters should not, but apart from that, all should
#' # be included.
#' transNames <- colnames(fullPanelUnmixed)[seq(6,18)]
#'
#' # ow, transform this file, with the default transformation factor of 400.
#' # NB! It is alway advisable to visually (or computationally) check the data
#' # for the most optimal transformation factors. These often vary from marker
#' # to marker.
#'
#' fullPanelTrans <- arcTrans(fullPanelUnmixed, transNames)
#'
#' @export arcTrans
arcTrans <- function(flowObj, transNames, transCoFacs = "default",
                     unTrans = FALSE) {
    if (inherits(flowObj, "flowSet")) {
        focusFrame <- flowObj[[1]]
    } else if (inherits(flowObj, "flowFrame")) {
        focusFrame <- flowObj
    } else {
        stop("The flowObj needs to be either a flowSet or a flowFrame")
    }

    if (length(transCoFacs) == 1) {
        if (transCoFacs == "default") {
            transCoFacs <- massOrFlowTrans(
                focusFrame = focusFrame,
                transNames = transNames
            )
        } else if (is.numeric(transCoFacs)) {
            transCoFacs <- rep(transCoFacs, length(transNames))
            names(transCoFacs) <- transNames
        }
    } else {
        names(transCoFacs) <- transNames
    }

    if (inherits(flowObj, "flowSet")) {
        resultObj <- fsApply(flowObj, arcTransCoFunction,
            transCoFacs = transCoFacs,
            unTrans = unTrans
        )
    } else {
        resultObj <- arcTransCoFunction(
            focusFrame = flowObj,
            transCoFacs = transCoFacs,
            unTrans = unTrans
        )
    }
    return(resultObj)
}

arcTransCoFunction <- function(focusFrame, transCoFacs, unTrans) {
    focusFrameResult <- focusFrame
    if (unTrans) {
        for (i in names(transCoFacs)) {
            exprs(focusFrameResult)[, i] <-
                sinh(exprs(focusFrame)[, i]) * transCoFacs[i]
        }
    } else {
        for (i in names(transCoFacs)) {
            exprs(focusFrameResult)[, i] <-
                asinh(exprs(focusFrame)[, i] / transCoFacs[i])
        }
    }

    message("Transformation complete")

    return(focusFrameResult)
}

# arcTrans
context("arcTrans")
value <- 5
coFactor <- 3
result <- asinh(value/coFactor)
valueFrame <- flowCore::flowFrame(matrix(value, 1, 1, dimnames = list(NA, "Val")))
transFrame <- arcTrans(valueFrame, transNames = colnames(valueFrame),
                       transCoFacs = coFactor)
test_that("arcTransResult", {
    expect_equal(as.numeric(exprs(transFrame)), result)
})

#Untransform the data again
unTransFrame <- arcTrans(transFrame, transNames = colnames(valueFrame),
                         transCoFacs = coFactor, unTrans = TRUE)
test_that("arcUnTransResult", {
    expect_equal(as.numeric(exprs(unTransFrame)), value)
})

# This function is a slightly adjusted copy of the fr_append_cols
# function in flowCore. It mainly allows for the presence of additional
# pData variables
# @param focusFrame The flowFrame that will get appended.
# @param newCols The new columns that should be added.
# @return The focusFrame, with the newCols appended.
#' @importFrom flowCore pData exprs parameters
appendFFCols <- function(focusFrame, newCols) {
    pd <- flowCore::pData(flowCore::parameters(focusFrame))
    cn <- colnames(newCols)
    new_pid <- max(as.integer(gsub("\\$P", "", rownames(pd)))) +
        1
    new_pid <- seq(new_pid, length.out = ncol(newCols))
    new_pid <- paste0("$P", new_pid)

    new_pd <- do.call(rbind, lapply(cn, function(i) {
        vec <- newCols[, i]
        rg <- range(vec)
        new_pd <- pd[1, ]
        new_pd[1, ] <- NA
        new_pd$name <- i
        new_pd$range <- diff(rg) + 1
        new_pd$minRange <- rg[1]
        new_pd$maxRange <- rg[2]
        return(new_pd)
    }))
    rownames(new_pd) <- new_pid
    pd <- rbind(pd, new_pd)
    newExprs(focusFrame) <- cbind(exprs(focusFrame), newCols)
    flowCore::pData(flowCore::parameters(focusFrame)) <- pd
    focusFrame
}

#' Generate a correction matrix for cytometry data analysis
#'
#'
#' This function aids the correctUnmix function, to create a symmetrical
#' correction matrix that should be used together with a flowframe to correct
#' the errors of unmixing.
#' @param specMat The spectral matrix used to unmix the dataset of interest.
#' @return A symmetrical matrix of zeros with the right row- and column names.
#' @seealso \code{\link{correctUnmix}}
#' @importFrom BiocGenerics ncol colnames
#' @examples
#' # Load uncompensated data
#' data(fullPanel)
#'
#' # Load the spectral unmixing matrix generated with controls from the same
#' # experiment. These can be generated using the specMatCalc function.
#' data(specMat)
#'
#' # And now unmix
#' fullPanelUnmix <- specUnmix(fullPanel, specMat)
#'
#'
#' # Create an empty unmixinng matrix
#' corrMat <- corrMatCreate(specMat)
#'
#' @export corrMatCreate
corrMatCreate <- function(specMat) {
    corrNames <- rownames(specMat)
    correcMat <- matrix(0,
        nrow = length(corrNames),
        ncol = length(corrNames),
        dimnames =
            list(corrNames, corrNames)
    )
    diag(correcMat) <- 1
    return(correcMat)
}

# Select events based on filter
#
#
# @param flowObj The fcs object to be subsetted. Both flowFrames and flowSets
# are accepted.
# @param filterName The name of the filter variable in the flow object to be
# used for the subsetting.
# @param withinFilter Should the events within or outsitde of the filter be
# selected?
# @return The flowObj, now smaller, as it only contains the data within,
# or outside of the filterName filter.
#' @importFrom flowCore fsApply
filterOut <- function(flowObj, filterName, withinFilter = TRUE) {
    if (withinFilter) {
        gateVal <- 1
    } else {
        gateVal <- 0
    }
    if (inherits(flowObj, "flowSet")) {
        resultObj <- fsApply(flowObj, function(x) {
            return(x[which(exprs(x[, filterName])[, 1] == gateVal), ])
        })
    } else if (inherits(flowObj, "flowFrame")) {
        resultObj <- flowObj[which(exprs(flowObj[, filterName])[, 1] ==
            gateVal), ]
    } else {
        stop("The flowObj needs to be either a flowSet or a flowFrame")
    }
    return(resultObj)
}

# Median absolute deviation-based automatic filters for flowCore objects
#
# This function is meant to create conservative, MAD-based filters for n-peaked
# cytometry variables. The function is so far not flowCore-compliant, and thus
# not available as a direct user function, but only used internally. This
# will change soon.
# @param flowObj The fcs object to be filtered. Both flowFrames and flowSets
# are accepted.
# @param gateVar The variable that should be used to set the gate. Can either
# be an integer or a string.
# @param nMads The number of median absolute deviations that should be
# included. These are calculated separately from each half of the peak, so
# if applied on both sides, there can still be an assymetry present.
# @param filterName The common name to the filter(s) created by the function.
# Default is the name of the gate variable.
# @param nGates The number of gates that should be produced
# @param madSide Which side of the peak(s) should the MAD filter be applied to?
# "low", "high", "both" and "none" supported.
# @param nonMadFilter What filter should be applied on the possible non-mad
# side? The three alternatives are:
# \describe{
#    \item{"deflection"}{Here, the gate is extended to the deflection point
#    marking the start of the next peak.} "none" and "default".
#    \item{"none"}{Here, all events are included}
# }
# It is worth noting that madSide overrides this argument.
# @param adjust The value deciding the accuracy of the density calculation. The
# higher the value, the lower the sensitivity for small aberrations in the
# density.
# @param returnSepFilter Should the gate be returned as a separate object?
# Currently, this defaults to FALSE. Incompatible with returnGateVals = TRUE.
# @param returnGateVals Should only the gate value(s) be returned? Currently, this
# defaults to FALSE. Incompatible with returnSepFilter = TRUE.
# @return If returnSepFilter and returnGateVals are both FALSE, flowObject of
# the same class as flowObj with the gates added as boolean variables to the
# exprs portions of the flowFrames. Otherwise, either a matrix containing the
# information on which cells that belong to whcih gate, or just a vector
# of gate values.
madFilter <- function(flowObj, gateVar = 1, nMads = 2, filterName = "default",
                      nGates = 1, madSide = "both", adjust = 2,
                      nonMadFilter = "deflection", returnSepFilter = FALSE,
                      returnGateVals = FALSE ) {
    # First, the gateVar is converted to an integer, if specified as a string

    if (madSide == "none" && nonMadFilter == "none") {
        stop("With these settings, no filtering would be achieved")
    }

    if (nGates > 1 && madSide != "both" && nonMadFilter == "none") {
        stop("The combination of multiple gates with no nonMadFilter would lead
             to overlapping gates")
    }

    if (returnSepFilter && returnGateVals){
        stop("Either the gate value or the result of applying this value",
             " is returned, not both")
    }

    if (is.character(gateVar)) {
        gateVar <- which(BiocGenerics::colnames(flowObj) == gateVar)
    }

    if (filterName == "default") {
        filterName <- paste0(
            BiocGenerics::colnames(flowObj)[gateVar],
            "_auto_filter"
        )
    }

    if (inherits(flowObj, "flowSet")) {
        resultObj <- fsApply(flowObj, madFilterCoFunction,
            gateVar = gateVar, nMads = nMads,
            filterName = filterName,
            nGates = nGates, madSide = madSide,
            nonMadFilter = nonMadFilter,
            adjust = adjust,
            returnSepFilter = returnSepFilter,
            returnGateVals = returnGateVals
        )
    } else if (inherits(flowObj, "flowFrame")) {
        resultObj <- madFilterCoFunction(flowObj,
            gateVar = gateVar,
            nMads = nMads,
            filterName = filterName,
            nGates = nGates, madSide = madSide,
            nonMadFilter = nonMadFilter,
            adjust = adjust,
            returnSepFilter = returnSepFilter,
            returnGateVals = returnGateVals
        )
    } else {
        stop("The flowObj needs to be either a flowSet or a flowFrame")
    }

    return(resultObj)
}

madFilterCoFunction <- function(focusFrame, gateVar, nMads,
                                filterName, nGates,
                                madSide, nonMadFilter,
                                adjust, returnSepFilter, returnGateVals) {
    focusVar <- exprs(focusFrame[, gateVar])[, 1]

    peakPlaces <- peakIdenti(focusVar,
        nPeaks = nGates,
        adjust = adjust, returnStats = TRUE
    )

    lowMads <- as.list(rep(NA, length = length(peakPlaces[[1]])))
    highMads <- lowMads

    if (madSide == "both" || madSide == "low") {
        lowMads <- lapply(seq_along(peakPlaces[[1]]), function(x)
            peakPlaces[[1]][x] - (
                peakMadCalc(focusVar[which(focusVar >= peakPlaces[[2]][[x]][1] &
                    focusVar < peakPlaces[[1]][x])],
                peakVal = peakPlaces[[1]][x]
                ) * nMads))
    }
    if (madSide == "both" || madSide == "high") {
        highMads <- lapply(seq_along(peakPlaces[[1]]), function(x)
            peakPlaces[[1]][x] + (
                peakMadCalc(focusVar[which(focusVar >= peakPlaces[[1]][x] &
                    focusVar <
                        peakPlaces[[2]][[x]][2])],
                peakVal = peakPlaces[[1]][x]
                ) * nMads))
    }

    # And now, the gates are created, according to the settings above.


    gateValList <- lapply(seq_along(peakPlaces[[1]]), function(x)
        madGateValsCreation(
            focusVar = focusVar, lowMad = lowMads[[x]],
            highMad = highMads[[x]], lowPeakEnd = peakPlaces[[2]][[x]][1],
            highPeakEnd = peakPlaces[[2]][[x]][2], madSide = madSide,
            nonMadFilter = nonMadFilter
        ))
    #And this list is exported, if the interest lies mainly in the gate values
    #and not in transforming the whole dataset
    if (returnGateVals) {
        return(gateValList)
    }
    #If we are interested in getting the actual information about the cells,
    #we get that information here.
    gateVecList <- lapply(seq_along(peakPlaces[[1]]), function(x){
        locVec <- rep(0, times = length(focusVar))
        locVec[which(focusVar >= gateValList[[x]][1] &
                         focusVar < gateValList[[x]][2])] <- 1
        locVec
    }
    )

    if (length(peakPlaces[[1]]) > 1) {
        gateVecMat <- do.call("cbind", gateVecList)

        colnames(gateVecMat) <-
            paste0(filterName, "_", seq_along(peakPlaces[[1]]))
    } else {
        gateVecMat <- matrix(unlist(gateVecList))
        colnames(gateVecMat) <- filterName
    }

    #Here, we export if we do not want the data to be integrated into the
    #flowset
    if (returnSepFilter) {
        return(gateVecMat)
    }
    #If we do want the data to be integrated in the flowSet, here is the place
    #to do so
    focusFrame <- appendFFCols(focusFrame, gateVecMat)
}

peakMadCalc <- function(focusHalfPeak, peakVal) {
    focusHalfPeakCent <- focusHalfPeak - peakVal
    focusPeak <- c(focusHalfPeakCent, focusHalfPeakCent * -1)

    return(mad(focusPeak))
}

#madVecCreation <- function(focusVar, lowMad, highMad, lowPeakEnd,
#                           highPeakEnd, madSide, nonMadFilter) {
#    resultVar <- rep(0, times = length(focusVar))
#    if (madSide == "both") {
#        resultVar[which(focusVar >= lowMad & focusVar < highMad)] <- 1
#    } else if (nonMadFilter == "deflection") {
#        if (madSide == "low") {
#            resultVar[which(focusVar >= lowMad & focusVar < highPeakEnd)] <- 1
#        } else if (madSide == "high") {
#            resultVar[which(focusVar >= lowPeakEnd & focusVar < highMad)] <- 1
#        } else {
#            resultVar[which(focusVar >= lowPeakEnd & focusVar < highPeakEnd)] <-
#                1
#        }
#    } else {
#        if (madSide == "low") {
#            resultVar[which(focusVar >= lowMad)] <- 1
#        } else if (madSide == "high") {
#            resultVar[which(focusVar < highMad)] <- 1
#        }
#    }
#    return(resultVar)
#}

madGateValsCreation <- function(focusVar, lowMad, highMad, lowPeakEnd,
                           highPeakEnd, madSide, nonMadFilter) {
    if (madSide == "both") {
        c(lowMad,highMad)
    } else if (nonMadFilter == "deflection") {
        if (madSide == "low") {
            c(lowMad,highPeakEnd)
        } else if (madSide == "high") {
            c(lowPeakEnd,highMad)
        } else {
            c(lowPeakEnd,highPeakEnd)
        }
    } else {
        if (madSide == "low") {
            c(lowMad,max(focusVar))
        } else if (madSide == "high") {
            c(min(focusVar),highMad)
        }
    }
}

#' Plotting all variables against a single variable
#'
#'
#' This function is useful both when setting appropriate gates and when the
#' adjustments of the compensation are done
#' @param flowObj This is the full dataset, either a flowFrame or a flowSet,
#' that should be plotted. If it has more rows than "nRows", a subsample (with
#' equal contributions from each sample if a flowSet) will be plotted.
#' @param yCol Here, the variable to be plotted against all the others is
#' selected. It can be either a number, the column name of interest or "all".
#' @param nRows The number of rows that will be used to construct the plot.
#' The fewer, the faster, but the resolution also decreases, naturally. Default
#' is 100000.
#' @param plotName If a name different from yCol should be used, it can be added
#' here.
#' @param dirName Add a custom directory name.
#' @param zeroTrim In the case of CyTOF data, the events at zero can often
#' be so dominant, that all other density variation is dwarfed, and thus
#' invisible. With this command, the events that are zero in both y
#' and x are trimmed for each x separately.
#' @param hexBins How many bins should the hexagonal plots be divided in?
#' @param saveResult Should the result be saved as a file?
#' @return A plot with one 2D-graph for each variable that the y-variable
#' should be plotted against.
#' @seealso \code{\link{correctUnmix}}
#' @importFrom flowCore fsApply exprs
#' @importFrom reshape2 melt
#' @importFrom ggplot2 ggplot aes facet_wrap geom_hex xlab ylab theme
#' element_blank element_line ggsave scale_fill_gradientn
#' @importFrom BiocParallel bplapply
#' @examples
#' #' # Load uncompensated data
#' data(fullPanel)
#'
#' # Load the spectral unmixing matrix generated with controls from the same
#' # experiment. These can be generated using the specMatCalc function.
#' data(specMat)
#'
#' # Now unmix
#' fullPanelUnmix <- specUnmix(fullPanel, specMat)
#'
#' # Transform all fluorescent channels
#' fullPanelTrans <- arcTrans(fullPanelUnmix,
#'     transNames = colnames(fullPanelUnmix)[6:18])
#'
#' # And now run the function. If no specific marker is selected, as in this
#' # case, then all markers will be plotted in a new sub-directory.
#' # Further, if you leave the saveResult to TRUE, a pdf will be created.
#' oneVsAllPlot(fullPanelTrans, yCol = "BV650_CD56", saveResult = FALSE)
#'
#' # This shows that there is a compensation artifact between AF647_IgM and
#' # BV650_CD56, which is an expected combination to cause problems, due to the
#' # similar emission characteristics. It is therefore recommended to go on to
#' # correctUnmix function.
#'
#' @export oneVsAllPlot
oneVsAllPlot <- function(flowObj, yCol = "all", nRows = 10000,
                         plotName = "default", dirName = "All_vs_all_plots",
                         zeroTrim = TRUE, hexBins = 30, saveResult = TRUE) {
    plotExprs <- plotDownSample(flowObj, nRows)
    
    if (yCol == "all") {
        dir.create(dirName)
        bplapply(seq_along(colnames(plotExprs)), function(x) {
            oneVsAllPlotCoFunction(
                plotExprs = plotExprs, yCol = x,
                plotName = plotName,
                zeroTrim = zeroTrim,
                hexBins = hexBins,
                saveResult = saveResult,
                dirName = dirName
            )
        })
    } else {
        oneVsAllPlotCoFunction(
            plotExprs = plotExprs, yCol = yCol,
            plotName = plotName,
            zeroTrim = zeroTrim,
            hexBins = hexBins,
            saveResult = saveResult
        )
    }
}

oneVsAllPlotCoFunction <- function(plotExprs, yCol, plotName, zeroTrim, hexBins,
                                   saveResult, dirName = ".") {
    # If the yCol is given as a character, it is converted to the correct number
    # here
    if (is.character(yCol)) {
        yCol <- which(colnames(plotExprs) == yCol)
    }
    
    if (plotName == "default") {
        plotName <- file.path(dirName, colnames(plotExprs)[yCol])
    }
    
    # Here, a separate y-name is created
    yName <- colnames(plotExprs)[yCol]
    
    # Now, the data is truncated, to minimize influence by individual events
    # on the display
    plotExprs <- apply(plotExprs, 2, function(x) {
        high <- quantile(x, 0.999)
        low <- quantile(x, 0.001)
        x[x > high] <- high
        x[x < low] <- low
        return(x)
    })
    
    # Here, the dataset is converted into a long format.
    longPlot <- melt(as.data.frame(plotExprs))
    longPlot$Y <- rep(plotExprs[, yCol], times = ncol(plotExprs))
    
    # Here, the double negative events are made NA for each x separately
    if (zeroTrim == TRUE) {
        longPlot[which(longPlot$value == 0 & longPlot$Y == 0), 2:3] <- c(NA, NA)
    }
    
    # Now, the plots are created.
    # Before this, we are just formally defining vale and Y, for the sake of
    # notes in CMD check.
    value <- 0
    Y <- 0
    p <- ggplot(longPlot, aes(x = value, y = Y)) +
        facet_wrap(~variable, scales = "free") +
        geom_hex(na.rm = TRUE, bins = hexBins) +
        scale_fill_gradientn(colours = c(
            "#440154FF", "#38598CFF", "#2D708EFF",
            "#25858EFF", "#1E9B8AFF", "#2BB07FFF",
            "#51C56AFF", "#85D54AFF", "#C2DF23FF",
            "#FDE725FF", "#FDE725FF", "#FDE725FF"
        )) +
        xlab("") + ylab(yName) + theme(
            axis.text = element_blank(),
            axis.ticks = element_blank(),
            panel.grid.major = element_blank(),
            panel.grid.minor = element_blank(),
            panel.background = element_blank(),
            axis.line = element_line(colour = "black")
        )
    if (saveResult) {
        ggsave(paste0(plotName, "_vs_all_others.pdf"))
    } else {
        print(p)
    }
}

#' Convert a flowSet to one long dataframe with all identifiers as separate
#' #columns.
#'
#'
#' This function is mainly used for compatibility with matrix-based clustering
#' algorithms, such as depeche in the DepecheR package.
#'
#' @importFrom flowCore fsApply exprs flowSet sampleNames keyword
#' @param flowObj The flowSet or flowFrame to be converted to a dataframe.
#' @param idInfo A list of any number of characteristics that can be derived
#' from the file names. For each entry, a gsub specification of where to find
#' the information in the file name should be added, such as id=""..._|..."".
#' @return A long data frame with one column per PMT/APD (or fluorochrome,
#' depending on the state of the imported files), one for the acquisition date
#' (for fcs files) and one colum for each specified slot above. If no
#' gsub-pattern is provided, only a single column with the full file name will
#' be used to separate the observations from each file.
#' @seealso \code{\link[DepecheR]{depeche}}
#' @examples
#' #' # Load uncompensated data
#' data(fullPanel)
#'
#' # Load the spectral unmixing matrix generated with controls from the same
#' # experiment. These can be generated using the specMatCalc function.
#' data(specMat)
#'
#' # Now unmix
#' fullPanelUnmix <- specUnmix(fullPanel, specMat)
#'
#' # Transform all fluorescent channels
#' fullPanelTrans <- arcTrans(fullPanelUnmix,
#'     transNames = colnames(fullPanelUnmix)[6:18])
#'
#' # This function is primarily meant to be used with flowSets.
#' # If we had only one flowFrame, we could just extract the data by
#' # the use of the flowCore function exprs(), so we will convert the data to a
#' # flowSet now.
#' library(flowCore)
#' fullPanelFs <- flowSet(fullPanelTrans)
#'
#' # Before converting to a dataframe it is important to get an idea of the
#' # structure of the names, to be able to extract meaningful parts of the name.
#' # Here, we have an exceptional case again, as the flowSet has just been
#' # created, so there is actually no meaningful name of the flowFrame inside
#' # it. So for example reasons, we will give it one now:
#' sampleNames(fullPanelFs) <- "PBMC_full_panel_d1.fcs"
#'
#' # And now, we generate the dataframe:
#' fullPanelDf <- flowSet2LongDf(fullPanelFs, idInfo =
#'                                  list("Tissue" = "|_full_panel_..\\.fcs",
#'                                      "Donor" = "...._full_panel_|\\.fcs"))
#' # This is the result
#' str(fullPanelDf)
#'
#' @export flowSet2LongDf
flowSet2LongDf <- function(flowObj, idInfo) {
    if (inherits(flowObj, "flowFrame")) {
        flowObj <- flowSet(flowObj)
    }
    flowSetExprs <- data.frame(fsApply(flowObj, exprs))
    nameVector <- sampleNames(flowObj)

    if (missing(idInfo)) {
        flowSetExprs$names <-
            unlist(retrieveFlowSetNames(
                nameVector = nameVector,
                specFlowSet = flowObj, gsubpattern = ""
            ))
    } else {
        for (i in seq_along(idInfo)) {
            flowSetExprs$names <-
                unlist(retrieveFlowSetNames(
                    nameVector = nameVector,
                    specFlowSet = flowObj,
                    gsubpattern = idInfo[[i]]
                ))
            colnames(flowSetExprs)[which(colnames(flowSetExprs) == "names")] <-
                names(idInfo)[i]
        }
    }

    dateVector <- as.vector(fsApply(
        flowObj,
        function(x) keyword(x)$`$DATE`
    ))
    flowSetExprs$acqDate <-
        unlist(retrieveFlowSetNames(
            nameVector = dateVector,
            specFlowSet = flowObj,
            gsubpattern = ""
        ))

    return(flowSetExprs)
}

retrieveFlowSetNames <- function(nameVector, specFlowSet, gsubpattern) {
    lengthList <- as.list(fsApply(specFlowSet, function(x) nrow(exprs(x))))
    nameVectorShort <- as.list(gsub(pattern = gsubpattern, "\\1", nameVector))
    return(as.vector(mapply(
        function(x, y) rep(x, times = y), nameVectorShort,
        lengthList
    )))
}

# correctUnmix
context("flowSet2LongDf")
library(flowCore)
data(fullPanel)

testFrame <- fullPanel[seq(1,2),1]
exprs(testFrame)[,1] <- c(1,2)
testSet <- flowSet(testFrame[1,], testFrame[2,])
sampleNames(testSet) <- c("Test_1.fcs", "Test_2.fcs")
testDf <- flowSet2LongDf(testSet, idInfo = list("Number" = "Test_|\\.fcs"))

test_that("longDfRes1.1", {
    expect_equal(1, testDf[1,1])
})

test_that("longDfRes1.2", {
    expect_equal("1", testDf[1,2])
})

test_that("longDfRes2.1", {
    expect_equal(2, testDf[2,1])
})

test_that("longDfRes2.2", {
    expect_equal("2", testDf[2,2])
})

test_that("longDfResDate", {
    expect_equal("25-Oct-2018", testDf[1,3])
})

#' Tools for processing high-dimensional cytometry files
#'
#' @description This package is intended to fill the role of conventional
#' cytometry pre-processing software, for spectral decomposition,
#' transformation, visualization and cleanup, and to aid further downstream
#' analyses, such as with DepecheR, by enabling transformation of flowFrames
#' and flowSets to dataframes. Functions for flowCore-compliant automatic
#' 1D-gating/filtering are in the pipe line.
#' It is worth noting here that even if there are dedicated spectral cytometers,
#' it is possible to increase the separation of the fluorochromes in a
#' conventional flow cytometer too, by just keeping all non-used channels open.
#' That will however also require the use of spectral unmixing, rather than
#' compensation, as the compensation functions generally require the
#' compensation matrix to be symmetrical. So please open all channels, and use
#' this software!
#'
#' @name flowSpecs-package
#' @aliases flowSpecs-package flowSpecs
#' @docType package
#' @author
#' Maintainer: Jakob Theorell <jakob.theorell@@ndcn.ox.ac.uk>
#' @seealso \code{\link[flowCore:flowCore-package]{flowCore}}
#' @keywords package
#' @importFrom stats density lsfit mad median quantile
NULL

#' Calculating the matrix used for spectral unmixing
#'
#'
#' This algoritm takes a flowSet containing single-stained controls and
#' negative controls, including an autofluorescence control and estimates the
#' unmixing for all fluorescent variables.
#' @param unmixCtrls A flowSet containing all the single stained and
#' unstained files necessary to create an spectral unmixing matrix. These can 
#' but do not have to, contain a negative control. Such a negative control will
#' not be used, and instead an universal negative control needs to be included
#' for each sample type present among the single-stained controls. 
#' @param groupNames A character vector containing strings common to the groups
#' of non-autofluoresence unmixCtrls that could be present. If for example
#' all antibodies single stains are anti-mouse bead-based the dead cell marker
#' is stained PBMC, and the files congruently either have a prefix containing
#' "Bead" or "PBMC", then the vector should be c("Bead", "PBMC"). The system
#' is not case specific.
#' @param autoFluoName The sample name of the autofluorescence control.
#' @return A data frame with each row representing a fluorochrome or
#' or autofluorescence and each column representing a detector.
#' @importFrom BiocGenerics colnames ncol
#' @importFrom stats var
#' @examples
#' # Load suitable unmixing controls. NB! If these originate from different
#' # sample types, such as beads and PBMC, there should be a negative control
#' # for each group and the names should reflect this, so that all PBMC samples
#' # would be called PBMC_unstained, PBMC_DCM, etc.
#' data(unmixCtrls)
#'
#' # If the dataset contains cell controls, make sure that the cell population
#' # interest dominates FSC-A, as the data highest peak in this channel will be
#' # used.
#'
#' # And run the function
#' specMat <- specMatCalc(unmixCtrls, groupNames = c("Beads_", "Dead_"),
#' autoFluoName = "PBMC_unstained.fcs")
#' @export specMatCalc
specMatCalc <- function(unmixCtrls, groupNames, autoFluoName) {


    # The spectrum for each file is calculated

    specCalcMat <- fsApply(unmixCtrls, specCalc)

    # Now, the samples are categorized into groups depending on their sample
    # type reflected in the names of the samples. If any samples are singlets,
    # then they are put to the side. The most likely reason for having one
    # singlet is that all unmixing controls have been acquired with beads,
    # but that the autofluorescence control is unstained cells.

    singleStainGroupsList <- lapply(groupNames, function(x)
        return(specCalcMat[which(grepl(x, row.names(specCalcMat))), ]))

    # Now, it is checked that if there is only one group, this group contains
    # more than samples, to make sure that unmixing is not attempted with one
    # color only, as this leads to some downstream negative effects, and also
    # is more or less meaningless.

    if(length(singleStainGroupsList) == 1 &&
       nrow(singleStainGroupsList[[1]]) < 3){
           stop("It seems like unmixing of one color is attempted, which is
           not meaningful. Please try again with another set of controls,
           or name them differently, if you believe that you have more than
           one color. ")
       }


    # Now, in each matrix in the list, the row with the lowest sum
    # is identified as the unstained
    negCtrlRows <- lapply(
        singleStainGroupsList,
        function(x) which.min(rowSums(x))
    )

    # If the autoFluoName is not in the

    # Here, the subtractions are made
    rawSpecMatList <- lapply(seq_along(negCtrlRows), function(x) {
        localSpecMat <- apply(singleStainGroupsList[[x]], 1, function(y)
            y - singleStainGroupsList[[x]][negCtrlRows[[x]], ])
    })

    # Here, the data is coerced into a matrix
    rawSpecMat <- do.call("cbind", rawSpecMatList)

    # Now, the unstained controls are removed
    specMatNoUnstain <- rawSpecMat[, -which(colSums(rawSpecMat) == 0)]

    # Here, the column names are cleaned up.
    specMatColNamesRaw1 <- colnames(specMatNoUnstain)
    specMatColNamesRaw2 <- gsub("|\\.fcs", "", specMatColNamesRaw1)
    specMatColNames <- vector()
    for (i in specMatColNamesRaw2) {
        for (j in groupNames) {
            if (grepl(j, i)) {
                specMatColNames[i] <- gsub(paste0(j, "|"), "", i)
            }
        }
    }

    colnames(specMatNoUnstain) <- specMatColNames

    # Now, the autofluorescence medians are added
    specMat <- cbind(specMatNoUnstain,
        "Autofluo" = specCalcMat[autoFluoName, ]
    )

    specMatFrac <- t(apply(specMat, 2, function(x) x / max(x)))

    # And finally, all negative values resulting from minor errors in detection,
    # are removed.
    specMatFrac[which(specMatFrac < 0)] <- 0

    return(specMatFrac)
}

specCalc <- function(flowFrame) {
    focusColNames <- BiocGenerics::colnames(flowFrame)

    # First, a gate is applied to FSC.A, to simplify work with cells
    fscVar <- which(grepl("FSC", focusColNames) &
        grepl("A", focusColNames))[1]

    fscGateVals <- madFilter(flowFrame, gateVar = fscVar, nMads = 1.5,
                               returnGateVals = TRUE)[[1]]
    fscVarDat <- exprs(flowFrame)[,fscVar]
    fscFilteredFrame <- flowFrame[which(fscVarDat > fscGateVals[1] &
                                            fscVarDat < fscGateVals[2]),]

    # Now, a similar gate is applied to ssc, to clean up all files.
    sscVar <- which(grepl("SSC", focusColNames) &
                        grepl("A", focusColNames))[1]
    
    sscGateVals <- madFilter(fscFilteredFrame, gateVar = sscVar, nMads = 1.5,
                             returnGateVals = TRUE)[[1]]
    sscVarDat <- exprs(fscFilteredFrame)[,sscVar]
    sscFilteredFrame <- fscFilteredFrame[which(sscVarDat > sscGateVals[1] &
                                            sscVarDat < sscGateVals[2]),]

    # Here, all non-fluorescent channels are excluded
    fluoFrame <- sscFilteredFrame[, -grep("ime|SC|ort|ate|omp", focusColNames)]
    
    #Now, if two peaks are present in any channels, then the top peak will
    #be selected. 
    topVar <- which.max(apply(exprs(fluoFrame), 2, var))
    
    transDat <- arcTrans(fluoFrame, colnames(fluoFrame)[topVar])
    
    #Are two peaks present?
    nPeaks <- peakIdenti(exprs(transDat[,topVar]))
    
    #If so, only the top peak events are used
    if(length(nPeaks) == 2){
        gateVals <- madFilter(transDat, topVar, nGates = 2, 
                              returnGateVals = TRUE)[[2]]
        fluoFrame <- fluoFrame[which(transDat[,topVar] > gateVals[1] &
                                         transDat[,topVar] < gateVals[2]),]
    }
    # Then the median is calculated for all fluorescence channels on this
    # filtered population
    fluoColNames <- BiocGenerics::colnames(fluoFrame)

    rawMedVals <- vapply(fluoColNames, function(x)
        median(exprs(fluoFrame[, x])), 1)


    # Now, the highest peak is identified, and the data further gated on this
    # variable, to reduce the variance further
    maxMedVar <- which.max(rawMedVals)

    # This channel is now produced separately, to increase computational
    # speed
    maxMedVarFrame <- fluoFrame[, maxMedVar]

    # And here, this channel is transformed for the madFilter to work correctly
    maxMedVarFrameTrans <- arcTrans(maxMedVarFrame,
        transNames = colnames(maxMedVarFrame),
        transCoFacs = 400
    )

    # Here, the data is gated
    maxGatedFrame <- madFilter(maxMedVarFrameTrans,
        gateVar = 1,
        nMads = 1.5, returnSepFilter = TRUE
    )
    maxFilteredFrame <- fluoFrame[which(maxGatedFrame == 1), ]

    # And finally, the median procedure is repeated for this final population
    resultMedVals <- vapply(fluoColNames, function(x)
        median(exprs(maxFilteredFrame[, x])), 1)
}

# correctUnmix
context("specMatCalc")
data(unmixCtrls)
testUnmixSet <- unmixCtrls[c(1,2,12,15)]

testSpecMat <- specMatCalc(testUnmixSet, groupNames = c("Beads_"),
                       autoFluoName = "PBMC_unstained.fcs")

test_that("massOrFlowTransResult", {
    expect_equal(1, testSpecMat[1,"R2-A"])
})

# The point of this function is to enable autodetection of the "optimal" value
# for the arcsinh cofactor in the transformation of the data. This is because
# CyTOF and flow cytometry data have very different needs, but the data can also
# easily be differentiated, as CyTOF data is much sparser.
# focusFrame = the frame that will be transformed.
# transNames = the names of the columns in the focusFrame that should be
# transformed.

massOrFlowTrans <- function(focusFrame, transNames) {
    if (unname(sum(exprs(focusFrame) == 0) / (dim(focusFrame)[1] *
        dim(focusFrame)[2])) > 0.05) {
        transCoFacs <- rep(5, times = length(transNames))
    } else {
        transCoFacs <- rep(400, times = length(transNames))
    }
    names(transCoFacs) <- transNames

    return(transCoFacs)
}

# correctUnmix
context("massOrFlowTrans")
data(fullPanel)

testFrame <- flowFrame(matrix(c(rep(0, 50), 1),
                              nrow= 51, ncol = 1, dimnames  =
                                  list(rep(NA, 51), "V1")))
as.numeric(flowSpecs:::massOrFlowTrans(testFrame, transNames = "V1"))

test_that("massOrFlowTransResult", {
    expect_equal(5, as.numeric(
        flowSpecs:::massOrFlowTrans(testFrame, transNames = "V1")))
})

#' Normalize batch differences in intensities by aligning peaks
#'
#'
#' This function is intended to be used on standardized controls,
#' preferrably one standard that has been acquired with every batch.
#' This function needs to be applied separately for each batch, with the same
#' standard.
#'
#' @param fs The flowSet to be normalized
#' @param ctrlPos The position in the flowSet that contains the internal
#' control
#' @param standFF The external standard to which all batches should be
#' normalized
#' @param transNames If parameters are not transformed prior to running this,
#' internal transformation is necessary to identify peaks, so specify which
#' variables that need transformation here. The data for these variables
#' are untransformed at the end, so the data will have the same scale in and
#' out.
#' @param transCoFacs Also inherited from arcTrans. Low values (2-10) for CyTOF
#' data, high values (200-2000) for flow data, all depending on the number of
#' input channels.
#' @param volThresh The threshold for how small the area of a peak can be compared to
#' the largest peak, and still count.
#' @param normOrNot This needs to be a logical vector with the same length as the
#' number of columns in the fs and standFF. If it is known that a certain
#' variable should not be normalized, that information can be specified
#' here.
#' @return A normalized flowSet. In addition, output to console, to clarify
#' #for which markers the same number of peaks was detected in the control and
#' the standard, as normalization will only be applied for these.
#' @importFrom flowCore sampleNames<-
#' @examples
#' #Load uncompensated data and spectral matrix.
#' data(fullPanel)
#' data(specMat)
#'
#' # And now unmix
#' fullPanelUnmix <- specUnmix(fullPanel, specMat)
#'
#'  #Create a new file with the value 1000 added to all values
#'  library(flowCore)
#'  fullPanelPlus1000 <- flowFrame(exprs(fullPanelUnmix)+1000)
#' # Check how they differ.
#' range(exprs(fullPanelUnmix)[,1])
#' # 143 1187733
#' range(exprs(fullPanelPlus1000)[,1])
#' # 1143 1188733
#'
#' #Now normalize the new one to the old. NB! Here we will only
#' normPanel1000 <- peakNorm(flowSet(fullPanelPlus1000), 1, fullPanelUnmix,
#' transNames = colnames(fullPanelUnmix)[6:18], transCoFacs = 500)
#'
#' #And now check the new result
#' range(exprs(normPanel1000[[1]] )[,1])
#' # 143 1187733
#' @export peakNorm
#'
peakNorm <- function(fs, ctrlPos, standFF, transNames = FALSE, transCoFacs,
                     volThresh = 0.005, normOrNot = rep(TRUE, ncol(standFF))){

    #First, the data is centered to the median in the standFF or the ctrlPos
    standMed <- apply(exprs(standFF), 2, median)
    ctrlMed <- apply(exprs(fs[[ctrlPos]]), 2, median)
    standFFCent <- medCent(standFF, standMed)
    fsCent <- fsApply(fs, medCent, ctrlMed)
    message("median centering done")

    #Now we make a rough arcTrans of the data, to get peaks
    if(length(transNames) == 1 && transNames == FALSE){
        standTrans <- standFFCent
        fsTrans <- fsCent
    } else {
        standTrans <- arcTrans(standFFCent, transNames, transCoFacs)
        fsTrans <- arcTrans(fsCent, transNames, transCoFacs)
    }

    #And now we normalize data. First, we identify the values in the control
    #frame, and then we apply them to all values in the flowset.
    peakVals <- lapply(
        list(exprs(standTrans),
             exprs(fsTrans[[ctrlPos]])), function(x){
                 locRes <- apply(x, 2, function(y){
                     locPeak <- peakIdenti(y, volThresh,
                                                       returnStats = TRUE)
                     if(length(locPeak$PeakPos) == 1){
                         peakPoss <- 1
                     } else {
                         peakPoss <- c(1,2)
                     }
                     locPeak$Median <- unlist(lapply(peakPoss, function(z){
                         median(y[
                             which(y > locPeak$Width[[z]][1] &
                                       y < locPeak$Width[[z]][2])])
                     }))
                     locPeak
                 })
             })

    fsFinal <- flowSet(lapply(seq_along(fsTrans), function(x){
        peakNormCoFunc(fsTrans[[x]], peakVals[[2]], peakVals[[1]],
                       standMed, inFF = fs[[x]], normOrNot,
                       transNames, transCoFacs)
    }))
    sampleNames(fsFinal) <- sampleNames(fs)
    fsFinal
}
#########
medCent <- function(FF, cent){
    locExprs <- exprs(FF)
    locResMedCentExprs <-
        do.call("cbind", lapply(seq_len(ncol(locExprs)),
                                function(y){
                                    yRes <- locExprs[,y] - cent[y]
                                }))
    colnames(locResMedCentExprs) <- colnames(locExprs)
    exprs(FF) <- locResMedCentExprs
    return(FF)
}
#########
peakNormCoFunc <- function(FF, ctrlPeakVals, standPeakVals, standMed, inFF,
                           normOrNot, transNames, transCoFacs){
    locRes <- lapply(seq_len(ncol(exprs(FF))), function(x){
        locDat <- exprs(FF)[,x]
        ctrlPeaks <- ctrlPeakVals[[x]]
        standPeaks <- standPeakVals[[x]]
        peaksMatch <- TRUE
        if(length(ctrlPeaks$Median) == 1 &&
           length(standPeaks$Median) == 1){
            subtrDat <- locDat - ctrlPeaks$Median
            normDat <- subtrDat + standPeaks$Median

        } else if(length(ctrlPeaks$Median) == 2 &&
                  length(standPeaks$Median) == 2){
            subtrDat <- (locDat - ctrlPeaks$Median[1])/
                (ctrlPeaks$Median[2] - ctrlPeaks$Median[1])
            normDat <- (subtrDat * (standPeaks$Median[2] -
                                        standPeaks$Median[1])) +
                standPeaks$Median[1]

        } else {
            normDat <- locDat
            peaksMatch <- FALSE
        }
        return(list(normDat, peaksMatch))

    })

    locResMat <- do.call("cbind", lapply(locRes, "[[", 1))
    colnames(locResMat) <- colnames(FF)
    locNormOrNot <- unlist(lapply(locRes, "[[", 2))
    for(i in seq_along(normOrNot)){
        if(locNormOrNot[i] == FALSE){
            normOrNot[i] <- FALSE
        }
    }
    names(normOrNot) <- colnames(FF)
    message(paste0(names(normOrNot), "=", normOrNot, " "))

    exprs(FF) <- locResMat
    #Now we can untransform the data again and bring it back to its
    #original values.
    if(length(transNames) == 1 && transNames == FALSE){
        FFUntrans <- FF
    } else {
        FFUntrans <- arcTrans(FF, transNames, transCoFacs, unTrans = TRUE)
    }

    #And now finally we reverse the median subtraction from the first step.
    FFOrig <- medCent(FFUntrans, -standMed)

    #Now, for the columns where there was a mismatch in the number of peaks
    #between the standard and the internal control, the original, untouched
    #data is returned
    exprs(FFOrig)[,which(normOrNot == FALSE)] <-
        exprs(inFF)[,which(normOrNot == FALSE)]

    return(FFOrig)
}

data(fullPanel)
extract <- fullPanel[,c(2,20)]

extractPlus1000 <- flowFrame(exprs(extract)+1000)

#Now normalize the new one to the old.
normPanel1000 <- peakNorm(flowSet(extractPlus1000), 1, extract)

#And now check the new result

test_that("massOrFlowTransResult", {
    expect_equal(range(exprs(extract)), range(exprs(normPanel1000[[1]])))
})

#' This function lets us exchange a flowframes exprs portion to an unrelated one
#' It is solely meant to be used internally, as it is a strange practice.
#' @param x A flowFrame
#' @param value A matrix suitable to be an exprs object.
#' @return A new flowFrame with "value" as the exprs portion.
#' @importFrom methods setGeneric setMethod
#' @keywords internal
setGeneric("newExprs<-", function(x, value) standardGeneric("newExprs<-"))

setMethod("newExprs<-", "flowFrame", function(x, value) {
    x@exprs <- value
    x
})

#' Peak identification for higher-level functions.
#'
#' This function is primarily thought to be used internally to define peaks
#' in data.
#' One function is borrowed from package vulcan, namely the vulcan::densityauc,
#' which is neat, but the package is large and significantly increases the
#' installation time, and the importing is thus discarded.
#' @param markerData The data that the peaks should be identified for
#' @param volThresh The cutoff ratio of the volume for each secondary peak,
#' under which it is not considered to be a peak
#' @param distThresh The cutoff under which two peaks are considered one, as
#' they are too close to each other. This value between 0 and 1 corresponds to a
#' fraction from the 10th to the 90th percentile of the data range that the
#' peaks must be separated by to count. Defaults to 0.1 or 10 percent of the
#' distance.
#' @param adjust The value deciding the accuracy of the density calculation. The
#' higher the value, the lower the sensitivity for small aberrations in the
#' density.
#' @param nPeaks The number of peaks that should be exported. If n+1
#' fulfilling the volRatio criterion are found, the peaks most separated in
#' space are chosen.
#' @param returnStats Should the deflection points defining the peaks, the peak
#' hight and the lowest deflection point between the two most extreme peaks
#' be included in the export?
#' @return The information about the peaks in question. Depending on if
#' returnStats is TRUE or not and the number of peaks, it will change in
#' complexity.
#' @seealso \code{\link[vulcan]{densityauc}}
#' @keywords internal
#' @importFrom zoo rollmean
peakIdenti <- function(markerData, volThresh = 0.05, distThresh = 0.1,
                       adjust = 2, nPeaks = 2, returnStats = FALSE) {
    #First, we remove outliers, as they can have unintended effects
    #especially on the distThresh, as true clusters can look very close if a
    #few extreme values are included.
    quantile1_99 <- quantile(markerData, c(0.001, 0.999))
    markerData <- markerData[which(markerData > quantile1_99[1] &
                                       markerData < quantile1_99[2])]
    Da <- density(markerData, adjust = adjust, na.rm = TRUE)
    DeltaY <- diff(Da$y)
    Turns <- which(DeltaY[-1] * DeltaY[-length(DeltaY)] < 0) + 1

    # In the case that the peak is located at the lowest position
    # in the density Y vector, it is included in the turns vector here
    if (which.max(Da$y) == 1) {
        Turns <- c(1, Turns)
    }

    # Here, the top peak decides if the peaks are odd or even numbers in Turns.
    maxPeakPos <- which.max(Da$y[Turns])
    if (maxPeakPos %% 2 == 0) {
        allPeakMaxima <- Turns[seq_along(Turns) %% 2 == 0]
    } else {
        allPeakMaxima <- Turns[seq_along(Turns) %% 2 != 0]
    }
    # Either, there is only one peak, and in that case, no further analyses are
    # needed. If not, a set of functions will find which peaks to choose
    if (length(allPeakMaxima) == 1) {
        if (returnStats) {
            return(list(
                "PeakPos" = Da$x[allPeakMaxima],
                "Width" = list(c(min(markerData), max(markerData))),
                "DensVol" = NA,
                "Height" = NA, "LowVertex" = NA
            ))
        } else {
            return(Da$x[allPeakMaxima])
        }
    }

    # Now, the volume for each peak is calculated.
    # Each peak is defined as the region from the preceding to the succeding
    # deflection point
    densVols <- vector()
    peakRegVals <- list()
    peakHeights <- vector()
    for (i in seq_along(allPeakMaxima)) {
        focusTurnPos <- which(Turns == allPeakMaxima[i])
        if (focusTurnPos == 1) {
            peakRegion <- c(1, Turns[2])
        } else if (focusTurnPos == length(Turns)) {
            peakRegion <- c(Turns[focusTurnPos - 1], length(Da$x))
        } else {
            peakRegion <- c(Turns[focusTurnPos - 1], Turns[focusTurnPos + 1])
        }
        localPeakRegVals <- c(Da$x[peakRegion[1]], Da$x[peakRegion[2]])

        # Now calculate the volume of the region in question
        xt <- diff(Da$x[Da$x > localPeakRegVals[1] & Da$x <
            localPeakRegVals[2]])
        yt <- rollmean(
            Da$y[Da$x > localPeakRegVals[1] & Da$x <
                localPeakRegVals[2]],
            2
        )
        densVols[i] <- sum(xt * yt)
        peakRegVals[[i]] <- localPeakRegVals
        peakHeights[i] <- Da$y[allPeakMaxima[i]]
    }

    # Here, the volumes are sorted by size
    allPeakMaximaSorted <- allPeakMaxima[order(densVols, decreasing = TRUE)]
    peakRegValsSorted <- peakRegVals[order(densVols, decreasing = TRUE)]
    densVolsOrdered <- densVols[order(densVols, decreasing = TRUE)]
    peakHeightsSorted <- peakHeights[order(densVols, decreasing = TRUE)]

    # Now, all peaks with a volume smaller than volThresh are excluded
    peakMaximaReal <- allPeakMaximaSorted[which(densVolsOrdered > volThresh)]
    peakRegValsReal <- peakRegValsSorted[which(densVolsOrdered > volThresh)]
    densVolsReal <- densVolsOrdered[which(densVolsOrdered > volThresh)]
    peakHeightsReal <- peakHeightsSorted[which(densVolsOrdered > volThresh)]

    # Now, if this returns only one value, this value is returned here
    if (length(peakMaximaReal) == 1 || nPeaks == 1) {
        if (returnStats) {
            return(list(
                "PeakPos" = Da$x[peakMaximaReal[1]],
                "Width" = peakRegValsReal[1],
                "DensVol" = densVolsReal[1],
                "Height" = peakHeightsReal[1],
                "LowVertex" = NA
            ))
        } else {
            return(Da$x[peakMaximaReal][1])
        }
    }

    # Here, it is investigaded if the peaks are separated by more than
    # a set fraction of the total width of the data. This is done
    # iteratively, first considering the largest and the second largest
    # peak, then considering if the third peak is separated from both the
    # first and the second peak, and so on. The intention with this
    # algorithm is to avoid detecting a peak with a minor aberration at the top
    # to be detected as two peaks.

    threshDist <- length(Da$x) * distThresh
    peakMaximaFinal <- peakMaximaReal[1]
    j <- 2
    for (i in seq(2, length(peakMaximaReal))) {
        if (all(vapply(
            seq(1, i - 1), function(x)
                abs(peakMaximaReal[x] - peakMaximaReal[i]) > threshDist,
            TRUE
        ))) {
            peakMaximaFinal[j] <- peakMaximaReal[i]
            j <- j + 1
        }
    }

    peakRegValsFinal <-
        peakRegValsReal[which(peakMaximaReal %in% peakMaximaFinal)]
    densVolsFinal <- densVolsReal[which(peakMaximaReal %in% peakMaximaFinal)]
    peakHeightsFinal <-
        peakHeightsReal[which(peakMaximaReal %in% peakMaximaFinal)]

    # Here, we are identifying the lowest vertex present between any of the
    # peaks
    dayPeakSection <- Da$y[seq(
        min(peakMaximaFinal),
        max(peakMaximaFinal)
    )]

    lowVertex <- dayPeakSection[which.min(dayPeakSection)]

    # Now, the exports of more complex solutions is generated
    if (nPeaks < length(peakMaximaFinal)) {
        mostSepVals <- vapply(peakMaximaFinal, function(x)
            sum(abs(x - peakMaximaFinal)), 1)

        reducedPositions <- sort(peakMaximaFinal[order(mostSepVals,
            decreasing = TRUE
        )][seq_len(nPeaks)])

        peakMaximaFinalReduced <- peakMaximaFinal[which(peakMaximaFinal
        %in% reducedPositions)]

        # And here comes the reduction of all the metadata
        peakRegValsFinalReduced <-
            peakRegValsFinal[which(peakMaximaFinal %in% reducedPositions)]
        densVolsFinalReduced <- densVolsFinal[which(peakMaximaFinal
        %in% reducedPositions)]
        peakHeightsFinalReduced <- peakHeightsFinal[
            which(peakMaximaFinal %in% reducedPositions)
        ]

        # Followed by sorting into the natural order
        peakMaximaFinalSorted <-
            peakMaximaFinalReduced[order(peakMaximaFinalReduced)]

        peakRegValsFinalSorted <-
            peakRegValsFinalReduced[order(peakMaximaFinalReduced)]

        densVolsFinalSorted <-
            densVolsFinalReduced[order(peakMaximaFinalReduced)]

        peakHeightsFinalSorted <-
            peakHeightsFinalReduced[order(peakMaximaFinalReduced)]
    } else {
        peakMaximaFinalSorted <- peakMaximaFinal[order(peakMaximaFinal)]
        peakRegValsFinalSorted <- peakRegValsFinal[order(peakMaximaFinal)]
        densVolsFinalSorted <- densVolsFinal[order(peakMaximaFinal)]
        peakHeightsFinalSorted <- peakHeightsFinal[order(peakMaximaFinal)]
    }

    if (returnStats) {
        return(list(
            "PeakPos" = Da$x[peakMaximaFinalSorted],
            "Width" = peakRegValsFinalSorted,
            "DensVol" = densVolsFinalSorted,
            "Height" = peakHeightsFinalSorted,
            "LowVertex" = lowVertex
        ))
    } else {
        return(Da$x[peakMaximaFinalSorted])
    }
}

#' A fully stained spectral cytometry sample
#'
#' This is a flowFrame with a PBMC sample stained with 12 fluorochromes.
#' Data acquired on a 44 detector, 3 laser Cytek Aurora® instrument by
#' J Theorell.
#' Date: 2018-10-25.
#'
#' @docType data
#'
#' @usage data(fullPanel)
#'
#' @format An object of class \code{"flowFrame"}
#'
#' @keywords datasets
"fullPanel"

#' Spectral unmixing matrix
#'
#' This matrix is generated using the specMatCalc function and the unmixCtrls
#' example file.
#'
#' @docType data
#'
#' @usage data(specMat)
#'
#' @format An object of class \code{"matrix"};
#'
#' @keywords datasets
"specMat"

#' Spectral unmixing of cytometry files
#'
#'
#' This function performs the central task of spectral unmixing, to convert the
#' raw photon detector input to "biological" proxy-signals.
#' @importFrom BiocGenerics colnames
#' @importFrom flowCore exprs exprs<- parameters parameters<-
#' @importFrom Biobase pData pData<-
#' @param flowObj The fcs object to be filtered. Both flowFrames and flowSets
#' are accepted.
#' @param specMat This is a matrix generated by the secMatCalc function,
#' possibly with edited row names.
#' @return The unmixed data. It will be returned in the format it was imported
#' as.
#'
#' @examples
#' # Load uncompensated data
#' data(fullPanel)
#'
#' # Load the spectral unmixing matrix generated with controls from the same
#' # experiment. These can be generated using the specMatCalc function.
#' data(specMat)
#'
#' # And now, just run the function
#' fullPanelUnmix <- specUnmix(fullPanel, specMat)
#' @export specUnmix
specUnmix <- function(flowObj, specMat) {
    if (inherits(flowObj, "flowSet")) {
        resultObj <- fsApply(flowObj, specUnmixCoFunction,
            specMat = specMat
        )
    } else if (inherits(flowObj, "flowFrame")) {
        resultObj <- specUnmixCoFunction(flowObj, specMat = specMat)
    }
    return(resultObj)
}

specUnmixCoFunction <- function(focusFrame, specMat) {
    fullExprs <- exprs(focusFrame)
    rawData <- fullExprs[, colnames(specMat)]
    # Make the least squares fit based on the raw, uncompensated data.
    ls_corr <- lsfit(x = t(specMat), y = t(rawData), intercept = FALSE)
    # Export the unmixed portion of the least squares result.
    unmixResult <- t(ls_corr$coefficients)
    #Now, insert the columns in their places, if they were there.
    #Otherwise, put the non-compensated ones first, and the compensated
    #ones after.
    if(length(which(row.names(specMat) %in% colnames(fullExprs))) > 1){
        fullExprs[, colnames(specMat)] <- unmixResult
        exprs(focusFrame) <- fullExprs
    } else {
        #In the case that we do spectral unmixing, a few things
        #need to be fiddled with to get the flowFrame right.
        newExprs(focusFrame) <-
            cbind(fullExprs[
                , -which(colnames(fullExprs) %in% colnames(specMat))],
                unmixResult)
        locParamDat <- pData(parameters(focusFrame))
        #Now, we separate the portions that are changed and unchanged.
        locParamDatOld <-
            locParamDat[-which(colnames(fullExprs) %in% colnames(specMat)),]
        locParamDatNew <-
            locParamDat[
                which(colnames(fullExprs) %in% colnames(specMat)),][
                    seq_len(ncol(unmixResult)),]
        locParamDatMerge <- rbind(locParamDatOld, locParamDatNew)
        locParamDatMerge$name <- colnames(exprs(focusFrame))
        locParamDatMerge$desc <- colnames(exprs(focusFrame))
        row.names(locParamDatMerge) <-
            paste0("$P", seq_along(colnames(exprs(focusFrame))))
        pData(parameters(focusFrame)) <- locParamDatMerge

    }

    return(focusFrame)
}

# correctUnmix
context("specUnmix")

testSpecMat <- matrix(c(1,0.5,0,1), ncol = 2, nrow = 2,
                      dimnames = list(c("Res1", "Res2"),
                                      c("Start1", "Start2")))
testFlowFrame <- flowCore::flowFrame(matrix(c(1,1,1,1),
                                            nrow = 2, ncol = 2,
                                            dimnames = list(c(NA, NA),
                                                            c("Start1",
                                                              "Start2"))))
resultExprs <- exprs(specUnmix(testFlowFrame, testSpecMat))

test_that("specUnmixResult1", {
    expect_equal("Res1", colnames(resultExprs)[1])
})

test_that("specUnmixResult2", {
    expect_equal(0.5, resultExprs[1,1])
})

test_that("specUnmixResult3", {
    expect_equal(1, resultExprs[1,2])
})

# This function is a co-function to the oneVsAllPlot function, to reduce
# computational time.
# @param flowObj The object to downsample
# @param nRows The number of rows in the downsampled sample.
# @return The downsampled flowObj.
#' @importFrom BiocGenerics nrow
plotDownSample <- function(flowObj, nRows = 10000) {
    if (inherits(flowObj, "flowSet")) {
        if (sum(fsApply(flowObj, BiocGenerics::nrow)) > nRows) {
            numIncluded <- floor(nRows / length(flowObj))
            plotExprs <- fsApply(flowObj, function(x)
                return(exprs(x)[sample(
                    seq_len(BiocGenerics::nrow(x)),
                    numIncluded
                ), ]))
        } else {
            plotExprs <- fsApply(flowObj, function(x) return(exprs(x)))
        }
    } else if (BiocGenerics::nrow(flowObj) > nRows) {
        plotExprs <-
            exprs(flowObj)[sample(
                seq_len(BiocGenerics::nrow(flowObj)),
                nRows
            ), ]
    } else {
        plotExprs <- exprs(flowObj)
        message(
            "The number of rows in the dataset was only ",
            nrow(plotExprs),
            ", so the output will be restricted to this"
        )
    }
    return(plotExprs)
}

# correctUnmix
context("correctUnmix")
valueFrame <-
    flowCore::flowFrame(matrix(c(1,1, 1, 1), 2, 2,
                     dimnames = list(c(NA, NA), c("V1", "V2"))))
corrMat <- matrix(c(1, -0.5, 0, 1), nrow = 2, ncol = 2,
                  dimnames = list(c("V1", "V2"), c("V1", "V2")))
locRes <- as.numeric(exprs(correctUnmix(
    valueFrame, corrMat, transCoFacs = FALSE))[1,1])
test_that("correctUnmixResult", {
    expect_equal(locRes, 1.5)
})

---
title: Example workflow for processing of raw spectral cytometry files
author: 
- name: Jakob Theorell
  affiliation: 
  - Oxford Autoimmune Neurology Group, Nuffield Department of Clinical Neurosciences, University of Oxford, Oxford, United Kingdom
  - Department of Clinical Neurosciences, Karolinska Institutet, Stockholm, Sweden
date: "`r Sys.Date()`"
vignette: >
  %\VignetteIndexEntry{Example workflow for processing of raw spectral cytometry files}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}    
output: 
  BiocStyle::html_document:
    titlecaps: false
    toc_float: true
---

```{r style, echo=FALSE, results='hide', message=FALSE}
library(BiocStyle)
library(knitr)
library(flowCore)
opts_chunk$set(error=FALSE, message=FALSE, warning=FALSE)
knitr::opts_chunk$set(echo = TRUE)
```

# Introduction

In this document, I aim at showing a typical analysis of a spectral cytometry
file, including the construction of the spectral decomposition matrix, the
actual decomposition, correction of the resulting file (as there generally
are minor differences between the single-stained controls and the fully
stained sample) and finally converting the resulting flowFrame or flowSet to a 
dataframe that can be used for any downstream application. 
Note: This whole package is very much dependent on 
**flowCore**, and much of the functionality herein works as an extention of the
basic **flowCore** functionality. 

# Installation
This is how to install the package, if that has not already been done: 
```{r, eval = FALSE}
if(!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")
BiocManager::install("flowSpecs")
```

# Example data description

The dataset that is used in this vinjette, and that is the example dataset in 
the package generally, is a PBMC sample stained with 12 fluorochrome-conjugated
antibodies against a wide range of leukocyte antigens. Included is also a set
of single-stained controls, that fill the same function with spectral cytometry
as in conventional ditto. The files were generated on a 44 channel Cytek Aurora®
instrument 2018-10-25. 
```{r}
library(flowSpecs)
library(flowCore)
data("unmixCtrls")
unmixCtrls
data('fullPanel')
fullPanel[,seq(4,7)]
```

As can be noted, flowSpecs adheres to flowCore standards, and thus uses 
flowFrames and flowSets as input to all user functions. 


# Construction of spectral unmixing matrix
To do this, we need the single-stained unmixing controls. As the fluorescent
sources can be of different kinds, such as from antibodies, fluorescent 
proteins, or dead cell markers, the specMatCalc function accepts any number 
of different such groups. However, the groups need to have a common part of
their names. If this was not the case during acquisition, the names of the fcs
files can always be changed afterwards. To check the names, run the 
sampleNames function from flowCore: 
```{r}
sampleNames(unmixCtrls)
```

This shows that we have three groups of samples: "Beads", "Dead" and "PBMC". The
two first are groups that define the fluorochromes from antibodies and the dead 
cell marker (which is pacific orange-NHS in this case). The last one, "PBMC", 
will be used for autofluorescence correction. For obvious reasons, the autofluo
control should always be from the same type of sample as the samples that will
be analyzed downstream. 
With this knowledge about the groups of samples, we can now create the matrix: 

```{r}
specMat <- specMatCalc(unmixCtrls, groupNames = c("Beads_", "Dead_"), 
                        autoFluoName = "PBMC_unstained.fcs")
str(specMat)
```
Here we can see that a matrix with the original fluorescence detector names
as column names, and the new fluorochrome/marker names as row names has been 
created. The function does a lot of preprocessing, with automatic gating of the
most dominant population, as well as picking out the positive population if a 
negative control was included in each sample, etc, to ensure the best possible 
resolution and consistency in the determination of the matrix. 

# Spectral unmixing

Now it is time to apply the newly constructed specMat to the fully stained 
sample. This is done in the following way: 

```{r}
fullPanelUnmix <- specUnmix(fullPanel, specMat)
fullPanelUnmix
```
Notable is that the names now have been exchanged for the fluorescent molecules
instead of the detector channels. The algorithm below this function is currently
least squares regression. 

# Transformation
As with all cytometry data, for correct interpretation, the data needs to be 
transformed using one of the lin-log functions. As the arcsinh function  is 
widely used and also has a single co-function that controls the level of 
compression aroud zero, it is used in this package. The function has a number
of built-in features, such as automatic detection of if the file comes from
mass or flow cytometry, and will give differenc cofactors accordingly. It is
however always the best practice to set the cofactors individually, to ensure
that no artifactual populations are created, which can happen, if there is 
too much resolution around zero. One automated strategy for this, which 
would make the arcTrans function unnecessary, is to use the [flowVS](https://www.bioconductor.org/packages/release/bioc/html/flowVS.html) package.

The arcTrans function requires the names of the variables that should be 
transformed to be specified.

```{r}
fullPanelTrans <- arcTrans(fullPanelUnmix, transNames = 
                            colnames(fullPanelUnmix)[6:18])
par(mfrow=c(1,2))
hist(exprs(fullPanelUnmix)[,7], main = "Pre transformation", 
     xlab = "AF700_CD4", breaks = 200)
hist(exprs(fullPanelTrans)[,7], main = "Post transformation", 
     xlab = "AF700_CD4", breaks = 200)
```
As can be seen in the histograms, the ranges, scales and resolution have now 
changed dramatically. 
(Biologically, the three peaks correspond to CD4- cells, CD4+myeloid cells and 
CD4+T-cells, respectively). 

# Investigation of possible unmixing artifacts 

An important step in the early processing of cytometry files is to investigate
if, or rather where, unmixing artifacts have arisen. There are multiple reasons
for the occurrence of such artifacts, but listing them are outside of the scope
of this vinjette. In the package, there is one function that is well suited for
for this task, and that is the oneVsAllPlot function. When used without 
specifying a marker, the function will create a folder and save all possible
combinations of markers to that folder. Looking at them gives a good overview
of the data. In this case, for the vinjette purpose, I am only plotting one 
of the multi-graphs. 

```{r}
oneVsAllPlot(fullPanelTrans, "AF647_IgM", saveResult = FALSE)
```

This shows a typical artifact between BV650_CD56 and AF647_IgM: it is 
biologically extremely unlikely that the higher expression one sees of CD56, 
the more extremely below zero do the values become for IgM_AF647. 

# Correction of artifacts

Now to one of the more controversial subjects of cytometry, that rightly causes
alarm amongst anyone concerned about reproducibility: the correction of 
artifacts. When this is done aided by fluorescence-minus-one controls (and an
automated function with that purpose is being considered for this package), it 
is less controversial, but even without them, one can follow a few rules, to
increase the usefulness of the data. It is namely important to note, that if 
artifacts, of the kind we will now start to correct, are left in the data, then
they are likely to cause incorrect interpretation of the results: 
As a rule of thumb, one can assume negative correlations for 
single-positive markers (i.e. positive for x but negative for y) to always be
artifacts, as true populations below a negative population cannot exist. Strong
positive correlations are unlikely, but occur in biology, so caution and 
biological considerations should be taken before any corrections of such are 
attempted, but they should nonetheless be considered, as leaving them in can
cause harm. 

Normally, when correcting flow cytometry results, one just changes the 
compensation matrix. In this case, however, where the compensation matrix is not
symmetrical, that becomes a non-trivial affair. For that reason, this package
introduces a correction matrix, which is a secondary, symmetric matrix only 
meant to be used on already unmixed files. It can for that reason take both
positive and negative values. 

When starting the correction phase, we have to create an empty correction
matrix. 

```{r}
corrMat <- corrMatCreate(specMat)
```

This is how this correction matrix is meant to be used: 
A value of 1 corresponds to +100% correction, a value of 0, to 0% correction. 
Thus: if the value 1 is added to the coordinate [x,y], then if event 1 has a 
value of 50 in marker x, then event 1 will get +50 in marker y. 

Practically, in our case, we see that it seems like BV650_CD56 is slightly 
"overunmixed" from AF647_IgM. This means that we should add a negative 
correction. Let us start with 0.1, or 10%. 

```{r}
corrMat["BV650_CD56", "AF647_IgM"] <- -0.1
fullPanelCorr <- correctUnmix(fullPanelUnmix, corrMat)
oneVsAllPlot(fullPanelCorr, "AF647_IgM", saveResult = FALSE)
```
Here, a few things can be noted. First, the correction function takes the non-
transformed file as input. Second, there is an automatic transformation within
this function, as it would be tedious, always having to rerun the arcTrans 
function during this phase, that is generally quite repetitive. 
Thrdly, we overdid it, as the population is now clearly "undermixed" instead,
with a considerable bleed-over of CD56+ cells into the IgM marker channel. 
Thus, we repeat it with a lower value. 

```{r}
corrMat["BV650_CD56", "AF647_IgM"] <- -0.03
fullPanelCorr <- correctUnmix(fullPanelUnmix, corrMat)
oneVsAllPlot(fullPanelCorr, "AF647_IgM", saveResult = FALSE)
```
This time, the result was satisfactory. There are other minor defects in the 
unmixing, however, such as between AF647_IgM and PE_X. This is typically the 
case, and as long as this needs to be done manually, it will use considerable
time, especialy for more complex panels. 
Notable is also that as we are just changing the correction matrix, and
redoing the analysis from the unmixed file every time, we do not need to take
the previous values into consideration. 

# Connecting to other non-flowCore compliant applications
Many clustering algorithms and similar take a matrix-like input. If data is
to be combined from multiple fcs files, and clustered together, the most 
convenient way might be to create a long data frame containing identifiers as
separate columns from the flowSet. The flowSpecs package contains a function to
do this. It works also for single flowFrames, but there it might be easier to
just extract the data with the exprs() function from flowCore. 

To set up our file for the task, we will convert it to a flowSet and change its
currently non-existent name to something useful. 

```{r}
fullPanelFs <- flowSet(fullPanelTrans)
sampleNames(fullPanelFs) <- "PBMC_full_panel_d1.fcs"
```

The function we are goning to use can chop up the name of the file into
multiple strings, if the right information is added in a gsub-compliant
format. These strings are then added as new columns to the resulting dataframe,
and if the fcs files have been systematically named (or their sampleNames 
changed to something systematic in accordance with the example above), we will
in this way be able to group the data based on the new categorizing columns.

```{r}
fullPanelDf <- flowSet2LongDf(fullPanelFs, idInfo =
        list("Tissue" = "|_full_panel_..\\.fcs", 
             "Donor" = "...._full_panel_|\\.fcs"))
str(fullPanelDf)
```
This dataframe can now be used in other applications. 

# Summary
In this vinjette, a typical spectral cytometry analysis is performed, which 
is currently the main objective with the package. However, a number of functions
for automatic gating, CyTOF fcs file cleanup, etc are in the pipe line and will
be added to the package in the coming months, together with new vinjettes. 

# Session information
```{r}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compCtrls.R
\docType{data}
\name{unmixCtrls}
\alias{unmixCtrls}
\title{Unmixing controls}
\format{
An object of class \code{"flowSet"}
}
\usage{
data(unmixCtrls)
}
\description{
This is a flowSet with 14 spectral unmixing controls:
11 single-stained bead populations, 1 unstained bead, one dead-cell-marker-
stained PBMC sample, one unstained PBMC sample, working as a control for the
dead cell marker, and one autofluorescence control, which is also unstained
PBMC (in fact the same sample as the negative control for the dead cell
marker). Data acquired on a 44 detector, 3 laser Cytek Aurora® instrument by
J Theorell.
Date: 2018-10-25.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/newExprs.R
\name{newExprs<-}
\alias{newExprs<-}
\title{This function lets us exchange a flowframes exprs portion to an unrelated one
It is solely meant to be used internally, as it is a strange practice.}
\usage{
newExprs(x) <- value
}
\arguments{
\item{x}{A flowFrame}

\item{value}{A matrix suitable to be an exprs object.}
}
\value{
A new flowFrame with "value" as the exprs portion.
}
\description{
This function lets us exchange a flowframes exprs portion to an unrelated one
It is solely meant to be used internally, as it is a strange practice.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oneVsAllPlot.R
\name{oneVsAllPlot}
\alias{oneVsAllPlot}
\title{Plotting all variables against a single variable}
\usage{
oneVsAllPlot(
  flowObj,
  yCol = "all",
  nRows = 10000,
  plotName = "default",
  dirName = "All_vs_all_plots",
  zeroTrim = TRUE,
  hexBins = 30,
  saveResult = TRUE
)
}
\arguments{
\item{flowObj}{This is the full dataset, either a flowFrame or a flowSet,
that should be plotted. If it has more rows than "nRows", a subsample (with
equal contributions from each sample if a flowSet) will be plotted.}

\item{yCol}{Here, the variable to be plotted against all the others is
selected. It can be either a number, the column name of interest or "all".}

\item{nRows}{The number of rows that will be used to construct the plot.
The fewer, the faster, but the resolution also decreases, naturally. Default
is 100000.}

\item{plotName}{If a name different from yCol should be used, it can be added
here.}

\item{dirName}{Add a custom directory name.}

\item{zeroTrim}{In the case of CyTOF data, the events at zero can often
be so dominant, that all other density variation is dwarfed, and thus
invisible. With this command, the events that are zero in both y
and x are trimmed for each x separately.}

\item{hexBins}{How many bins should the hexagonal plots be divided in?}

\item{saveResult}{Should the result be saved as a file?}
}
\value{
A plot with one 2D-graph for each variable that the y-variable
should be plotted against.
}
\description{
This function is useful both when setting appropriate gates and when the
adjustments of the compensation are done
}
\examples{
#' # Load uncompensated data
data(fullPanel)

# Load the spectral unmixing matrix generated with controls from the same
# experiment. These can be generated using the specMatCalc function.
data(specMat)

# Now unmix
fullPanelUnmix <- specUnmix(fullPanel, specMat)

# Transform all fluorescent channels
fullPanelTrans <- arcTrans(fullPanelUnmix,
    transNames = colnames(fullPanelUnmix)[6:18])

# And now run the function. If no specific marker is selected, as in this
# case, then all markers will be plotted in a new sub-directory.
# Further, if you leave the saveResult to TRUE, a pdf will be created.
oneVsAllPlot(fullPanelTrans, yCol = "BV650_CD56", saveResult = FALSE)

# This shows that there is a compensation artifact between AF647_IgM and
# BV650_CD56, which is an expected combination to cause problems, due to the
# similar emission characteristics. It is therefore recommended to go on to
# correctUnmix function.

}
\seealso{
\code{\link{correctUnmix}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/flowSet2LongDf.R
\name{flowSet2LongDf}
\alias{flowSet2LongDf}
\title{Convert a flowSet to one long dataframe with all identifiers as separate
#columns.}
\usage{
flowSet2LongDf(flowObj, idInfo)
}
\arguments{
\item{flowObj}{The flowSet or flowFrame to be converted to a dataframe.}

\item{idInfo}{A list of any number of characteristics that can be derived
from the file names. For each entry, a gsub specification of where to find
the information in the file name should be added, such as id=""..._|..."".}
}
\value{
A long data frame with one column per PMT/APD (or fluorochrome,
depending on the state of the imported files), one for the acquisition date
(for fcs files) and one colum for each specified slot above. If no
gsub-pattern is provided, only a single column with the full file name will
be used to separate the observations from each file.
}
\description{
This function is mainly used for compatibility with matrix-based clustering
algorithms, such as depeche in the DepecheR package.
}
\examples{
#' # Load uncompensated data
data(fullPanel)

# Load the spectral unmixing matrix generated with controls from the same
# experiment. These can be generated using the specMatCalc function.
data(specMat)

# Now unmix
fullPanelUnmix <- specUnmix(fullPanel, specMat)

# Transform all fluorescent channels
fullPanelTrans <- arcTrans(fullPanelUnmix,
    transNames = colnames(fullPanelUnmix)[6:18])

# This function is primarily meant to be used with flowSets.
# If we had only one flowFrame, we could just extract the data by
# the use of the flowCore function exprs(), so we will convert the data to a
# flowSet now.
library(flowCore)
fullPanelFs <- flowSet(fullPanelTrans)

# Before converting to a dataframe it is important to get an idea of the
# structure of the names, to be able to extract meaningful parts of the name.
# Here, we have an exceptional case again, as the flowSet has just been
# created, so there is actually no meaningful name of the flowFrame inside
# it. So for example reasons, we will give it one now:
sampleNames(fullPanelFs) <- "PBMC_full_panel_d1.fcs"

# And now, we generate the dataframe:
fullPanelDf <- flowSet2LongDf(fullPanelFs, idInfo =
                                 list("Tissue" = "|_full_panel_..\\\\.fcs",
                                     "Donor" = "...._full_panel_|\\\\.fcs"))
# This is the result
str(fullPanelDf)

}
\seealso{
\code{\link[DepecheR]{depeche}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/arcTrans.R
\name{arcTrans}
\alias{arcTrans}
\title{Efficient inverse hyperbolic cosine transformation}
\usage{
arcTrans(flowObj, transNames, transCoFacs = "default", unTrans = FALSE)
}
\arguments{
\item{flowObj}{The fcs object to be transformed. Both flowFrames and flowSets
are accepted.}

\item{transNames}{The variables that should be normalized.}

\item{transCoFacs}{This value or vector of values define the values for the
transformation during the normalization. In the "default" case, the function
defines the object as a CyTOF object if >5 percent of the values are 0, and
applies the transformation value 5. Otherwise, the value 400 is applied.}

\item{unTrans}{If the reverse action should be taken, i.e. if an already
transformed dataset should be un-transformed. NB! It is of great importance
that the same transformation factors are used!}
}
\value{
A flow object containing the transformed data, and with all metadata
left untouched.
}
\description{
This is a simple wrapper function for the base asinh function, that is useful
for flowFrames and flowSets. It also allows for reversing the transformation
with the argument "unTrans".
}
\examples{
# Import some data and the spectral matrix. The latter can be generated using
# specMatCalc
data(fullPanel)
data(specMat)

fullPanelUnmixed <- specUnmix(fullPanel, specMat)

# Identify the columns that should be transformed
colnames(fullPanelUnmixed)
# The time and scatter parameters should not, but apart from that, all should
# be included.
transNames <- colnames(fullPanelUnmixed)[seq(6,18)]

# ow, transform this file, with the default transformation factor of 400.
# NB! It is alway advisable to visually (or computationally) check the data
# for the most optimal transformation factors. These often vary from marker
# to marker.

fullPanelTrans <- arcTrans(fullPanelUnmixed, transNames)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/corrMatCreate.R
\name{corrMatCreate}
\alias{corrMatCreate}
\title{Generate a correction matrix for cytometry data analysis}
\usage{
corrMatCreate(specMat)
}
\arguments{
\item{specMat}{The spectral matrix used to unmix the dataset of interest.}
}
\value{
A symmetrical matrix of zeros with the right row- and column names.
}
\description{
This function aids the correctUnmix function, to create a symmetrical
correction matrix that should be used together with a flowframe to correct
the errors of unmixing.
}
\examples{
# Load uncompensated data
data(fullPanel)

# Load the spectral unmixing matrix generated with controls from the same
# experiment. These can be generated using the specMatCalc function.
data(specMat)

# And now unmix
fullPanelUnmix <- specUnmix(fullPanel, specMat)


# Create an empty unmixinng matrix
corrMat <- corrMatCreate(specMat)

}
\seealso{
\code{\link{correctUnmix}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fullPanel.R
\docType{data}
\name{fullPanel}
\alias{fullPanel}
\title{A fully stained spectral cytometry sample}
\format{
An object of class \code{"flowFrame"}
}
\usage{
data(fullPanel)
}
\description{
This is a flowFrame with a PBMC sample stained with 12 fluorochromes.
Data acquired on a 44 detector, 3 laser Cytek Aurora® instrument by
J Theorell.
Date: 2018-10-25.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/specUnmix.R
\name{specUnmix}
\alias{specUnmix}
\title{Spectral unmixing of cytometry files}
\usage{
specUnmix(flowObj, specMat)
}
\arguments{
\item{flowObj}{The fcs object to be filtered. Both flowFrames and flowSets
are accepted.}

\item{specMat}{This is a matrix generated by the secMatCalc function,
possibly with edited row names.}
}
\value{
The unmixed data. It will be returned in the format it was imported
as.
}
\description{
This function performs the central task of spectral unmixing, to convert the
raw photon detector input to "biological" proxy-signals.
}
\examples{
# Load uncompensated data
data(fullPanel)

# Load the spectral unmixing matrix generated with controls from the same
# experiment. These can be generated using the specMatCalc function.
data(specMat)

# And now, just run the function
fullPanelUnmix <- specUnmix(fullPanel, specMat)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/specMat.R
\docType{data}
\name{specMat}
\alias{specMat}
\title{Spectral unmixing matrix}
\format{
An object of class \code{"matrix"};
}
\usage{
data(specMat)
}
\description{
This matrix is generated using the specMatCalc function and the unmixCtrls
example file.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/correctUnmix.R
\name{correctUnmix}
\alias{correctUnmix}
\title{Correct defects in spectral unmixing by compensation}
\usage{
correctUnmix(unmixFlowObj, corrMat, transCoFacs = 400)
}
\arguments{
\item{unmixFlowObj}{A flowframe or flowset post unmixing.}

\item{corrMat}{A correction matrix. If this is the first round, the
executionof this function needs to be preceeded by the generation of this
matrix, for example by using the \code{\link{corrMatCreate}} function.}

\item{transCoFacs}{If transformation should be performed, the transformation
cofactors can be added here. Three possible inputs: a vector with specific
cofactors for each variable, a set value that will be used for all variables,
and FALSE.
Note: It might be good to set this to FALSE in the final round, to optimize
the transoformations externally.}
}
\value{
The unmixed flow object, now corrected with the values from the
corrMat.
}
\description{
This function provides a way to reduce the defects in the spectral unmixing,
by creating a secondary correction matrix, which is symmetrical.
}
\examples{
# Load uncompensated data
data(fullPanel)

# Load the spectral unmixing matrix generated with controls from the same
# experiment. These can be generated using the specMatCalc function.
data(specMat)

# And now unmix
fullPanelUnmix <- specUnmix(fullPanel, specMat)


# Create an empty unmixinng matrix
corrMat <- corrMatCreate(specMat)

# Now correct the data with this. In the first instance, this will of course
# not have any effect, more  than transformation, as the corrMat is empty.
fullPanelCorr <- correctUnmix(fullPanelUnmix, corrMat)

# This now needs to be investigated, to identify any possible compensation
# defects. This is most easily done with the oneVsAllPlot executed in the
# following way:
\dontrun{
oneVsAllPlot(fullPanelCorr)
}
# One obvoius defect that shows when doing this is between CD56 and IgM:
oneVsAllPlot(fullPanelCorr, "BV650_CD56", saveResult = FALSE)

# This is corrcted the following way:
corrMat["BV650_CD56", "AF647_IgM"] <- -0.03
fullPanelCorr <- correctUnmix(fullPanelUnmix, corrMat)
oneVsAllPlot(fullPanelCorr, "BV650_CD56", saveResult = FALSE)

# This process is iterated until there are no remaining artifacts. Good help
# to do this is a set of fluorescence-minus-one controls. If that is not
# available, a rule of thumb is that if the signal in marker x is
# strongly negatively correlated to marker y, so that highly
# single-x-posisive values are below zero, then this is with all likelihood
# an artifact. The situation becomes more complicated with strong positive
# correlations, as they can occur in biology, so there one has to take more
# care and keep the marker biology in mind.

}
\seealso{
\code{\link{specUnmix}}, \code{\link{arcTrans}},
\code{\link{corrMatCreate}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/peakNorm.R
\name{peakNorm}
\alias{peakNorm}
\title{Normalize batch differences in intensities by aligning peaks}
\usage{
peakNorm(
  fs,
  ctrlPos,
  standFF,
  transNames = FALSE,
  transCoFacs,
  volThresh = 0.005,
  normOrNot = rep(TRUE, ncol(standFF))
)
}
\arguments{
\item{fs}{The flowSet to be normalized}

\item{ctrlPos}{The position in the flowSet that contains the internal
control}

\item{standFF}{The external standard to which all batches should be
normalized}

\item{transNames}{If parameters are not transformed prior to running this,
internal transformation is necessary to identify peaks, so specify which
variables that need transformation here. The data for these variables
are untransformed at the end, so the data will have the same scale in and
out.}

\item{transCoFacs}{Also inherited from arcTrans. Low values (2-10) for CyTOF
data, high values (200-2000) for flow data, all depending on the number of
input channels.}

\item{volThresh}{The threshold for how small the area of a peak can be compared to
the largest peak, and still count.}

\item{normOrNot}{This needs to be a logical vector with the same length as the
number of columns in the fs and standFF. If it is known that a certain
variable should not be normalized, that information can be specified
here.}
}
\value{
A normalized flowSet. In addition, output to console, to clarify
#for which markers the same number of peaks was detected in the control and
the standard, as normalization will only be applied for these.
}
\description{
This function is intended to be used on standardized controls,
preferrably one standard that has been acquired with every batch.
This function needs to be applied separately for each batch, with the same
standard.
}
\examples{
#Load uncompensated data and spectral matrix.
data(fullPanel)
data(specMat)

# And now unmix
fullPanelUnmix <- specUnmix(fullPanel, specMat)

 #Create a new file with the value 1000 added to all values
 library(flowCore)
 fullPanelPlus1000 <- flowFrame(exprs(fullPanelUnmix)+1000)
# Check how they differ.
range(exprs(fullPanelUnmix)[,1])
# 143 1187733
range(exprs(fullPanelPlus1000)[,1])
# 1143 1188733

#Now normalize the new one to the old. NB! Here we will only
normPanel1000 <- peakNorm(flowSet(fullPanelPlus1000), 1, fullPanelUnmix,
transNames = colnames(fullPanelUnmix)[6:18], transCoFacs = 500)

#And now check the new result
range(exprs(normPanel1000[[1]] )[,1])
# 143 1187733
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/specMatCalc.R
\name{specMatCalc}
\alias{specMatCalc}
\title{Calculating the matrix used for spectral unmixing}
\usage{
specMatCalc(unmixCtrls, groupNames, autoFluoName)
}
\arguments{
\item{unmixCtrls}{A flowSet containing all the single stained and
unstained files necessary to create an spectral unmixing matrix. These can 
but do not have to, contain a negative control. Such a negative control will
not be used, and instead an universal negative control needs to be included
for each sample type present among the single-stained controls.}

\item{groupNames}{A character vector containing strings common to the groups
of non-autofluoresence unmixCtrls that could be present. If for example
all antibodies single stains are anti-mouse bead-based the dead cell marker
is stained PBMC, and the files congruently either have a prefix containing
"Bead" or "PBMC", then the vector should be c("Bead", "PBMC"). The system
is not case specific.}

\item{autoFluoName}{The sample name of the autofluorescence control.}
}
\value{
A data frame with each row representing a fluorochrome or
or autofluorescence and each column representing a detector.
}
\description{
This algoritm takes a flowSet containing single-stained controls and
negative controls, including an autofluorescence control and estimates the
unmixing for all fluorescent variables.
}
\examples{
# Load suitable unmixing controls. NB! If these originate from different
# sample types, such as beads and PBMC, there should be a negative control
# for each group and the names should reflect this, so that all PBMC samples
# would be called PBMC_unstained, PBMC_DCM, etc.
data(unmixCtrls)

# If the dataset contains cell controls, make sure that the cell population
# interest dominates FSC-A, as the data highest peak in this channel will be
# used.

# And run the function
specMat <- specMatCalc(unmixCtrls, groupNames = c("Beads_", "Dead_"),
autoFluoName = "PBMC_unstained.fcs")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/peakIdenti.R
\name{peakIdenti}
\alias{peakIdenti}
\title{Peak identification for higher-level functions.}
\usage{
peakIdenti(
  markerData,
  volThresh = 0.05,
  distThresh = 0.1,
  adjust = 2,
  nPeaks = 2,
  returnStats = FALSE
)
}
\arguments{
\item{markerData}{The data that the peaks should be identified for}

\item{volThresh}{The cutoff ratio of the volume for each secondary peak,
under which it is not considered to be a peak}

\item{distThresh}{The cutoff under which two peaks are considered one, as
they are too close to each other. This value between 0 and 1 corresponds to a
fraction from the 10th to the 90th percentile of the data range that the
peaks must be separated by to count. Defaults to 0.1 or 10 percent of the
distance.}

\item{adjust}{The value deciding the accuracy of the density calculation. The
higher the value, the lower the sensitivity for small aberrations in the
density.}

\item{nPeaks}{The number of peaks that should be exported. If n+1
fulfilling the volRatio criterion are found, the peaks most separated in
space are chosen.}

\item{returnStats}{Should the deflection points defining the peaks, the peak
hight and the lowest deflection point between the two most extreme peaks
be included in the export?}
}
\value{
The information about the peaks in question. Depending on if
returnStats is TRUE or not and the number of peaks, it will change in
complexity.
}
\description{
This function is primarily thought to be used internally to define peaks
in data.
One function is borrowed from package vulcan, namely the vulcan::densityauc,
which is neat, but the package is large and significantly increases the
installation time, and the importing is thus discarded.
}
\seealso{
\code{\link[vulcan]{densityauc}}
}
\keyword{internal}
